/*
Original by Julien
Heavily modified 2008, 2011 by Sven Killig <sven@killig.de>
*/
#include "4D.h"

_4DModule::_4DModule() : portHandle(0) {
 	screeninfo.id			 = "4DModule";
	screeninfo.xres			 = 0;
 	screeninfo.yres			 = 0;
	screeninfo.xres_virtual		 = 0;
	screeninfo.yres_virtual		 = 0;
	screeninfo.xoffset		 = 0;
	screeninfo.yoffset		 = 0;
	screeninfo.bits_per_pixel	 = 0;
	screeninfo.height		 = 0;
	screeninfo.width		 = 0;
	screeninfo.writeWaitForACK 	 = 1;
	screeninfo.timeout		 = 100;
	screeninfo.baud_rate	 	 = 115200;
	screeninfo.deviceName	 	 = NULL;
	screeninfo.hardwareVersion	 = 1;
	screeninfo.firmwareVersion	 = 1;
}

_4DModule::_4DModule(const char * deviceName) : portHandle(0) {
 	screeninfo.id			 = "4DModule";
	screeninfo.xres			 = 0;
 	screeninfo.yres			 = 0;
	screeninfo.xres_virtual		 = 0;
	screeninfo.yres_virtual		 = 0;
	screeninfo.xoffset		 = 0;
	screeninfo.yoffset		 = 0;
	screeninfo.bits_per_pixel	 = 0;
	screeninfo.height		 = 0;
	screeninfo.width		 = 0;
	screeninfo.writeWaitForACK 	 = 1;
	screeninfo.timeout		 = 100;
	screeninfo.baud_rate	 	 = 115200;
	screeninfo.deviceName	 	 = deviceName;
	screeninfo.hardwareVersion	 = 1;
	screeninfo.firmwareVersion	 = 1;
}

_4DModule::_4DModule(const char * deviceName, int baudrate) : portHandle(0) {
 	screeninfo.id			 = "4DModule";
	screeninfo.xres			 = 0;
 	screeninfo.yres			 = 0;
	screeninfo.xres_virtual		 = 0;
	screeninfo.yres_virtual		 = 0;
	screeninfo.xoffset		 = 0;
	screeninfo.yoffset		 = 0;
	screeninfo.bits_per_pixel	 = 0;
	screeninfo.height		 = 0;
	screeninfo.width		 = 0;
	screeninfo.writeWaitForACK 	 = 1;
	screeninfo.timeout		 = 100;
	screeninfo.baud_rate	 	 = baudrate;
	screeninfo.deviceName	 	 = deviceName;
	screeninfo.hardwareVersion	 = 1;
	screeninfo.firmwareVersion	 = 1;
}

/*Private method to open the serial port. As it is private we don't have to check if it's already opened or not.*/
/*It does not take any arguments, everything is sourced from our private descriptor screeninfo which the caller	*/
/*should have made sure it's initialised to proper values (ie =! from the value assigned by the constructor.	*/
bool _4DModule::open(void) {
	#if debug
		fprintf(stderr,"open(\n");
	#endif
	//The structure to hold termios attributes
	struct termios attributes;
	//Reminder: NEVER use O_NONBLOCK here. For non-blocking I/O poll has been implemented.
	if( ( portHandle = ::open(screeninfo.deviceName, O_RDWR/* || O_NONBLOCK*/) ) <= 0 ){
		#if debug
			fprintf( stderr, "Could not open serial port: %s\n", screeninfo.deviceName);
		#endif
		return 0;
	}
	if( ( tcgetattr(portHandle, &attributes) ) < 0){
		#if debug
			fprintf( stderr, "Could not get serial port attributes\n");
		#endif
		return 0;
	}
	//We want raw access to the port
	cfmakeraw(&attributes);
	//This effectively sets the baud rate. It would have been good to use termios2 to take maximum advantage
	//of serial port speed but it's not yet implemented for USB to serial devices and requires a recent kernel (2.6.22+)
	cfsetspeed(&attributes, screeninfo.baud_rate);
	if ( ( tcsetattr(portHandle, TCSANOW, &attributes ) ) < 0 ){
		#if debug
			fprintf(stderr, "Could not set serial attributes\n");
		#endif
	 	return 0;
	}
	tcflush(portHandle, TCIOFLUSH); // flush input and output buffers
	return 1;
}

/*Private close method. I prefer the user to use moddown so all necessary steps are done correctly.*/
void _4DModule::close(void) {
	if(this->isOpen()){
		::close(portHandle);
	}
	portHandle = 0;
}

/*Public method to check if the serial port is opened... or not.*/
bool _4DModule::isOpen(void) {
	if(portHandle)
		return 1;
	return 0;
}

/*Private method to change baudrate. I prefer users to use modconfig.*/
bool _4DModule::setBaudRate(int baudrate) {
	screeninfo.baud_rate = baudrate;					//Update the baudrate in our screen descriptor
	if(this->isOpen()) {							//If it's opened, we access the port directly
		struct termios attributes;
		if( ( tcgetattr(portHandle, &attributes) ) < 0 ) {
			#if debug
			fprintf( stderr, "could not get serial port attributes\n");
			#endif
		return 0;
		}
		cfsetspeed(&attributes, screeninfo.baud_rate);			//Effectively change the baudrate
fprintf(stderr,"CRTSCTS!\n");
attributes.c_cflag |= CRTSCTS;
		if ( ( tcsetattr(portHandle, TCSANOW, &attributes) ) < 0 ) {
			#if debug
			fprintf(stderr,"could not set serial attributes\n");
			#endif
	 	return 0;
		}
	}
	return 1;
}

void _4DModule::setTimeout(unsigned int time)
{
	screeninfo.timeout = time;
}

/*Private method and low level as well, to get an ACK from the module.*/
/*We basically wait ~100ms or an ACK from the module (default). Use   */
/*setTimeout to change this value.				      */
bool _4DModule::__getACK(void){

	int retval = 0;
	nfds_t nfd = 1;
	pollfd mypoll;
	mypoll.fd = portHandle;
	mypoll.events = POLLIN;

	retval = ::poll(&mypoll, nfd, screeninfo.timeout);
	if(retval <= 0) {
		#if debug
		fprintf(stderr, "!::poll()!\n");
		#endif
		return 0;
	}
	if( mypoll.revents == POLLIN ) {
		char ACK = 0;
		::read(portHandle, &ACK, 1);
		if( ACK == 0x06 ) {
			#if debug
//			fprintf(stderr, "GOT ACK\n");
			#endif
			return 1;
		} else if( ACK == 0x15 ) {
			#if debug
			    fprintf(stderr, "NACK!\n");
			#endif
			return 0;
		}
	}
	#if debug
	fprintf(stderr, "Timeout or ERROR\n");
	#endif
	return 0;
}

/*Private method to obtain informations about the screen and to store them in our private screen descriptor for later use*/
bool _4DModule::__getScreenInfo(void){
	char getVersion[] = {'V', 0};			//No display on screen. Use displayScreenInfo method for that.
	char answer[5];					//The module replies back with 5 bytes.
	if(::write(portHandle, &getVersion, 2) <= 0)
		return 0;
				//We don't need ACK here, the module replies directly if command successful.
/*	if(::read(portHandle, answer, 5) <= 0)
		return 0;*/
	if(::read(portHandle, answer+0, 1) <= 0) return 0;
	if(::read(portHandle, answer+1, 1) <= 0) return 0;
	if(::read(portHandle, answer+2, 1) <= 0) return 0;
	if(::read(portHandle, answer+3, 1) <= 0) return 0;
	if(::read(portHandle, answer+4, 1) <= 0) return 0;
	switch(answer[0]) {				//From 4D Users' manual
		case 0:
			screeninfo.id = "OLED";
			break;
		case 1:
			screeninfo.id = "LCD";
			break;
		case 2:
			screeninfo.id = "VGA";
			break;
		default:
			screeninfo.id = "4DModule";
			break;
	}
	screeninfo.hardwareVersion = answer[1];
	screeninfo.firmwareVersion = answer[2];
	switch(answer[3]) {
		case 0x22:
			screeninfo.xres = 220;
			break;
		case 0x24:
			screeninfo.xres = 240;
			break;
		case 0x28:
			screeninfo.xres = 128;
			break;
		case 0x32:
			screeninfo.xres = 320;
			break;
		case 0x60:
			screeninfo.xres = 160;
			break;
		case 0x64:
			screeninfo.xres = 64;
			break;
		case 0x76:
			screeninfo.xres = 176;
			break;
		case 0x96:
			screeninfo.xres = 96;
			break;
		default:
			break;
	}
	switch(answer[4]) {
		case 0x22:
			screeninfo.yres = 220;
			break;
		case 0x28:
			screeninfo.yres = 128;
			break;
		case 0x32:
			screeninfo.yres = 320;
			break;
		case 0x60:
			screeninfo.yres = 160;
			break;
		case 0x64:
			screeninfo.yres = 64;
			break;
		case 0x76:
			screeninfo.yres = 176;
			break;
		case 0x96:
			screeninfo.yres = 96;
			break;
		default:
			break;
	}
	return 1;
}

/*This method is public. It displays the informations about the screen directly onto the screen.*/
bool _4DModule::displayScreenInfo(void){
	if( ( this-isOpen() ) <= 0 )
		return 0;
	char showVersion[] = {'V', 1};
	if(::write(portHandle, &showVersion, 2) <= 0)
		return 0;
	return 1;
}

/*Our public methods to configure the class. NB: analogy to UNIX's ifconfig*/
bool _4DModule::modconfig(const char * deviceName) {
	screeninfo.deviceName = deviceName;
	if(this->isOpen()){				//If the serial port is already opened we close it
		close();				//because modconfig operates immediately;
		if(open() == 0)				//open the new one.
			return 0;
	}
	if(open() == 0)
		return 0;
	return 1;					//don't open a serial port if the previous one was closed
}
bool _4DModule::modconfig(const char * deviceName, int baudrate) {
	screeninfo.deviceName = deviceName;
	if(this->isOpen()) {
		close();
		if( setBaudRate(baudrate) && open() )
			return 1;
	}
	else {
		if( setBaudRate(baudrate) && open() )
			return 1;
	}
	return 0;
}
bool _4DModule::modconfig(int baudrate) {		//Here we don't need to mess with open/close
	if(setBaudRate(baudrate))			//since setBaudRate will take care of these
		return 1;				//and we did not touch deviceName;
	return 0;
}

/*Public method to bring up the module. Analogy to UNIX's ifup*/
bool _4DModule::modup(void) {
	#if debug
		fprintf(stderr,"modup(\n");
	#endif
	if(screeninfo.deviceName != NULL){						//Has been initialised
		/*if(open() && initialise() && eraseScreen() &&  __getScreenInfo()) {	//All-in-one
			return 1;
		}*/
		int o;
		o=open();
		if(o) {
			o=initialise(); // 'U'
			if(o) {
				/*o=eraseScreen();
				if(o) {*/
					o=__getScreenInfo();
					if(o) {
						return 1;
					} else {
					#if debug
						fprintf(stderr,"!__getScreenInfo()!\n");
					#endif
					}
				/*} else {
				#if debug
					fprintf(stderr,"!eraseScreen()!\n");
				#endif
				}*/
			} else
			{
			#if debug
				fprintf(stderr,"!initialise()!\n");
			#endif
			}
		} else
		{
		#if debug
			fprintf(stderr,"!open()!\n");
		#endif
		}
	}
	return 0;
}

/*Public method to bring down the module _safely_. Analogy to UNIX's ifdown*/
/*Here a return value of 0 just means that some steps failed. There is no  */
/*reason for close to fail and we wouldn't know either. So it's powerDown  */
/*which failed.								   */
bool _4DModule::moddown(void){
	if( ( this-isOpen() ) <= 0 )
		return 0;
	bool returnvalue = 1;
	if(powerDown() == 0)
		returnvalue = 0;
	close();
	return returnvalue;
}

/*Public for the moment. But is useless since users cannot call directly open. This method resend the 4D's modules*/
/*sync character: U, 0x55.											  */
int _4DModule::initialise(void){
	#if debug
		fprintf(stderr,"initialise(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char init = 'U';
	if( ::write(portHandle, &init, 1) <= 0 ) {
		return 0;
		#if debug
			fprintf(stderr,"!::write()!\n");
		#endif
	}
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*Public methods to write directly to the port (raw access)*/
/*They really are helpers functions so users can implement */
/*their own stuff on top of the library, eg PmmCLoader OSE */
/*use them to implement firmware downloading.		   */
int _4DModule::write(const char * buffer, int bytes, char pos1, char pos2, char pos3, char pos4){
  if(screeninfo.yres>255 || screeninfo.xres>255) {
  	return ::write(portHandle, buffer, bytes);
  } else {
  	int i;
        char send[1];
  	for(i=0;i<bytes;i++) {
  		if (i != pos1 && i != pos2 && i != pos3 && i != pos4) {
	  			send[0]=buffer[i];
	  			::write(portHandle, send, 1);
  		}
  	}
  	return i;
  }
}

int _4DModule::write(const char * buffer, int bytes){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	return ::write(portHandle, buffer, bytes);
}
int _4DModule::write(unsigned int * buffer, int bytes){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	return ::write(portHandle, buffer, bytes);
}
/*Public: those writes returns ACK from the module if any. Non-blocking*/
int _4DModule::writeWaitForACK(const char * buffer, int bytes){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	if(::write(portHandle, buffer, bytes) <= 0)
		return 0;
	return __getACK();
}
int _4DModule::writeWaitForACK(unsigned int * buffer, int bytes){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	if(::write(portHandle, buffer, bytes) <= 0)
		return 0;
	return __getACK();
}

/*Public: simple, blocking read. For asynchronous I/O use readTimeout.*/
int _4DModule::read(const char * buffer, int bytes){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	return ::read(portHandle, (void *)buffer, bytes);
}
/*Public: try to read from the serial port if possible. Timeout after screeninfo.timeout*/
/*Non-blocking.										*/
int _4DModule::readTimeout(const char * buffer, int bytes){
	int retval = 0;
	nfds_t nfd = 1;
	pollfd mypoll;
	mypoll.fd = portHandle;
	mypoll.events = POLLIN;

	retval = ::poll(&mypoll, nfd, screeninfo.timeout);
	if(retval <= 0)
		return 0;
	if( mypoll.revents == POLLIN ) {
		return ::read(portHandle, (void *)buffer, bytes);
	}
	return 0;
}

/*Public method to set the behaviour of the class toward ACK.*/
void _4DModule::writeWaitForACK(bool mode){
	screeninfo.writeWaitForACK = mode;
}

/*Public method which return our private screen descriptor for users' use.*/
var_screeninfo _4DModule::getScreenInfo(void){
	return screeninfo;
}

/*Public: resets the module :) */
bool _4DModule::reset(void){
	#if debug
		fprintf(stderr,"reset()\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	int flags;
	ioctl(portHandle, TIOCMGET, &flags);
	//clear DTR
	flags &= ~TIOCM_DTR;
	ioctl(portHandle, TIOCMSET, &flags);
	usleep(10);
	//Set DTR, issuing a RESET command
	flags |= TIOCM_DTR;
	ioctl(portHandle, TIOCMSET, &flags);
	usleep(100);
	return 1;
}

/*END of general methods. Follow: general command set.*/

bool _4DModule::addCharacter(unsigned char number, std::vector<unsigned char> * data){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	if( (number < 0) || (number > 31) )
		return 0;
	unsigned char cmd[] = {0x41, number};
	if(data->size() < 8) {
		char padd = 0;
		::write(portHandle, &cmd, 2);
		for(char i=0; i<8; i++){
			if(i < data->size()) {
				if(::write(portHandle, &data->at(i), 1) <= 0)
					return 0;
			}
			else {
				if(::write(portHandle, &padd, 1) <= 0)
					return 0;
			}
		}
	}
	else if(data->size() >= 8) {
		::write(portHandle, &cmd, 2);
		for(char i=0; i<8; i++){
			if(::write(portHandle, &data->at(i), 1) <= 0)
				return 0;
		}
	}
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*Public: set the background color of the module.*/
/*Arguments: a color coded on 2 bytes		 */
bool _4DModule::setBackground(short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x42, msb(color), lsb(color)};
	if( ::write(portHandle, &cmd, sizeof(cmd)) <= 0 )
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*Public: place a button on the screen.*/
/*Arguments: the state of the button, its position and color on the screen, and the text to be displayed on it.*/
bool _4DModule::placeButton(bool state, char x, short int y, short int color, char font, short int textColor, char textWidth, char textHeight,
			    std::string * buttonText) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x62, state, x, msb(y), lsb(y), msb(color), lsb(color), font, msb(textColor), lsb(textColor), textWidth, textHeight};
	char terminator = 0;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, buttonText->c_str(), buttonText->size()) <= 0)
		return 0;
	if(::write(portHandle, &terminator, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeButton(bool state, char x, short int y, short int color, char font, short int textColor, char textWidth, char textHeight,
			    const char * buttonText){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x62, state, x, msb(y), lsb(y), msb(color), lsb(color), font, msb(textColor), lsb(textColor), textWidth, textHeight};
	char terminator = 0;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, buttonText, strlen(buttonText)) <= 0)
		return 0;
	if(::write(portHandle, &terminator, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeButton(bool state, char x, char y, short int color, char font, short int textColor, char textWidth, char textHeight,
			    std::string * buttonText) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x62, state, x, y, msb(color), lsb(color), font, msb(textColor), lsb(textColor), textWidth, textHeight};
	char terminator = 0;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, buttonText->c_str(), buttonText->size()) <= 0)
		return 0;
	if(::write(portHandle, &terminator, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeButton(bool state, char x, char y, short int color, char font, short int textColor, char textWidth, char textHeight,
			    const char * buttonText){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x62, state, x, y, msb(color), lsb(color), font, msb(textColor), lsb(textColor), textWidth, textHeight};
	char terminator = 0;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, buttonText, strlen(buttonText)) <= 0)
		return 0;
	if(::write(portHandle, &terminator, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeButton(bool state, char x1, char y1, char x2, char y2, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x62, state, x1, y1, x2, y2, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}


bool _4DModule::drawCircle(short int x, short int y, short int radius, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'C', msb(x), lsb(x), msb(y), lsb(y), msb(radius), lsb(radius), msb(color), lsb(color)};
	if(_4DModule::write(cmd, sizeof(cmd), 1, 3, 5, -1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::drawCircleFill(char x, char y, char radius, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x69, x, y, radius, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::drawEllipse(char x, short int y, char rx, short int ry, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x65, x, msb(y), lsb(y), rx, msb(ry), lsb(ry), msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*bool _4DModule::copyPaste(char x, short int y, char xd, short int yd, char width, short int height){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x63, x, y, xd, msb(yd), lsb(yd), width, msb(height), lsb(height)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
// bool _4DModule::copyPaste(char xs, char ys, char xd, char yd, char width, char height){}
bool _4DModule::copyPaste(char x1, char y1, char x2, char y2, char x3, char y3){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'c', x1, y1, x2, y2, x3, y3};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}*/
bool _4DModule::copyPaste(short int xs, short int ys, short int xd, short int yd, short int width, short int height){//, char sPage, char dPage){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'c', msb(xs), lsb(xs), msb(ys), lsb(ys), msb(xd), lsb(xd), msb(yd), lsb(yd), msb(width), lsb(width), msb(height), lsb(height)};//, sPage, dPage};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}


bool _4DModule::displayCharacter(char group, char number, char x, short int y, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x44, group, number, x, msb(y), lsb(y), msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::displayCharacter(char number, char x, char y, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x44, number, x, y, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*Public: erase the screen.*/
bool _4DModule::eraseScreen(void){
	#if debug
		fprintf(stderr,"eraseScreen(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char erase = 0x45;
	if( ::write(portHandle, &erase, 1) <= 0 )
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::setFontSize(char size){
	#if debug
		fprintf(stderr,"setFontSize(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'F', size};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::drawTriangle(char x1, short int y1, char x2, short int y2, char x3, short int y3, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x47, x1, msb(y1), lsb(y1), x2, msb(y2), lsb(y2), x3, msb(y3), lsb(y3), msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::drawTriangle(char x1, char y1, char x2, char y2, char x3, char y3, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x47, x1, y1, x2, y2, x3, y3, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::drawPolygon(std::vector<short int> * vertices, unsigned int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	if( (vertices->size()/2 < 3) || (vertices->size()/2 > 7) ) //a vertice is defined by (x,y), so number of vertices is number of coordinates / 2
		return 0;
	char cmd[] = {0x67, vertices->size()/2};		//vertices->size() really should be even
	short int temp_msb, temp_lsb;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	for(char i=0; i<vertices->size()/2; i++) {
		temp_lsb = lsb(vertices->at(2*i));		//send x coordinate
		if(::write(portHandle, &temp_lsb, 1) <= 0)
			return 0;
		temp_msb = msb(vertices->at(2*i+1));		//prepare y coordinates
		temp_lsb = lsb(vertices->at(2*i+1));
		if(::write(portHandle, &temp_msb, 1) <= 0)	//send msb of y coordinates
			return 0;
		if(::write(portHandle, &temp_lsb, 1) <= 0)	//send lsb of y coordinates
			return 0;
	}
	temp_msb = msb(color);					//don't forget the color
	temp_lsb = lsb(color);
	if(::write(portHandle, &temp_msb, 1) <= 0)
		return 0;
	if(::write(portHandle, &temp_lsb, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::drawPolygon(std::vector<char> * vertices, unsigned int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	if( (vertices->size()/2 < 3) || (vertices->size()/2 > 7) )	//a vertice is defined by (x,y), so number of vertices is number of coordinates / 2
		return 0;
	char cmd[] = {0x67, vertices->size()/2};
	char temp_msb, temp_lsb;
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	for(char i=0; i<vertices->size(); i++) {
		if(::write(portHandle, &vertices->at(i), 1) <= 0)
			return 0;
	}
	temp_msb = msb(color);
	temp_lsb = lsb(color);
	if(::write(portHandle, &temp_msb, 1) <= 0)
		return 0;
	if(::write(portHandle, &temp_lsb, 1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::displayImage(unsigned short int x, unsigned short int y, unsigned short int width, unsigned short int height, char colorMode, std::vector<unsigned short int> * pixels){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'I', msb(x), lsb(x), msb(y), lsb(y), msb(width), lsb(width), msb(height), lsb(height), colorMode};
	int nbytes = height*width;
	char temp_msb, temp_lsb;
	if( !((colorMode == 16) || (colorMode == 8)) )
		return 0;
	_4DModule::write(cmd, sizeof(cmd), 1, 3, 5, 7);

	if(colorMode == 16) {
		for(int i=0; i < nbytes; i++) {
			temp_msb = msb(pixels->at(i));
			temp_lsb = lsb(pixels->at(i));
			if(::write(portHandle, &temp_msb, 1) <= 0)
				return 0;
			if(::write(portHandle, &temp_lsb, 1) <= 0)
				return 0;
		}
	}
	else {
		for(int i=0; i < nbytes; i++) {
			temp_lsb = lsb(pixels->at(i));
			if(::write(portHandle, &temp_lsb, 1) <= 0)
				return 0;
		}
	}


/*
int bytes=0;
char inbuff = 0;
if(screeninfo.firmwareVersion & 64) {
  if( (readTimeout(&inbuff, sizeof(inbuff))) <= 0 ){
  	std::cout << "Error\n";
  	if(inbuff!=0) std::cout << "(1:inbuff!=0)\n";
  	exit(0);
  }
}
for(int i=0; i < nbytes; i++) {
	if(colorMode == 16) {
		temp_msb = msb(pixels->at(i));
		bytes++;
		if(::write(portHandle, &temp_msb, 1) <= 0)
			return 0;
	}
	temp_lsb = lsb(pixels->at(i));
	bytes++;
	if(::write(portHandle, &temp_lsb, 1) <= 0)
		return 0;
      	if(bytes%128==0 && (screeninfo.firmwareVersion & 64)) {
//std::cout << i << " ";
      	        bytes=0;
      		if( (readTimeout(&inbuff, sizeof(inbuff))) <= 0 ){
      			std::cout << "Error at " << i << "\n";
      			if(inbuff!=0) std::cout << "(2:inbuff!=0)\n";
      			exit(0);
      		}
        }
}
*/

	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::drawLine(short int x1, short int y1, short int x2, short int y2, short int color){
	#if debug
		fprintf(stderr,"drawLine(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'L', msb(x1), lsb(x1), msb(y1), lsb(y1), msb(x2), lsb(x2), msb(y2), lsb(y2), msb(color), lsb(color)};
	if(_4DModule::write(cmd, sizeof(cmd), 1, 3, 5, 7) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	else
		usleep(10*1000);
	return 1;
}

bool _4DModule::setTextOpacity(bool mode){
	#if debug
		fprintf(stderr,"setTextOpacity(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'O', mode};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	else
		usleep(10*1000);
	return 1;
}

bool _4DModule::putPixel(short int x, short int y, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'P', msb(x), lsb(x), msb(y), lsb(y), msb(color), lsb(color)};
	if(_4DModule::write(cmd, sizeof(cmd), 1, 3, -1, -1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::setPenSize(char size){
	#if debug
		fprintf(stderr,"setPenSize(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'p', size};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

/*Public: reads a pixel's color.*/
int  _4DModule::readPixel(char x, short int y){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	int received;
	char cmd[] = {0x52, x, msb(y), lsb(y)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::read(portHandle, &received, 2) <= 0)
		return 0;
	return received;
}
int  _4DModule::readPixel(char x, char y){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	int received;
	char cmd[] = {0x52, x, y};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::read(portHandle, &received, 2) <= 0)
		return 0;
	return received;
}

bool _4DModule::drawRectangle(short int x1, short int y1, short int x2, short int y2, short int color){
	#if debug
		fprintf(stderr,"drawRectangle(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'r', msb(x1), lsb(x1), msb(y1), lsb(y1), msb(x2), lsb(x2), msb(y2), lsb(y2), msb(color), lsb(color)};
	if(_4DModule::write(cmd, sizeof(cmd), 1, 3, 5, 7) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	else
		usleep(100*1000);
	return 1;
}

bool _4DModule::paintArea(char x1, char y1, char x2, char y2, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x70, x1, y1, x2, y2, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::placeString(char x, short int y, char font, short int color, char width, char height, std::string * character){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x53, x, msb(y), lsb(y), font, msb(color), lsb(color), width, height};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character->c_str(), character->size()) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeString(char x, char y, char font, short int color, char width, char height, std::string * character){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x53, x, y, font, msb(color), lsb(color), width, height};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character->c_str(), character->size()) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeString(char column, char row, char font, short int color, std::string * character){
	#if debug
		fprintf(stderr,"placeString(%i, %i, %i, %i, std::string* '%s')\n", column, row, font, color, character->c_str());
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	//char cmd[] = {0x53, column, row, font, msb(color), lsb(color)};
	char cmd[] = {'s', column, row, font, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character->c_str(), character->size()+1) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	else
		usleep(100*1000);
	return 1;
}
bool _4DModule::placeString(char x, short int y, char font, short int color, char width, char height, const char * character){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x53, x, msb(y), lsb(y), font, msb(color), lsb(color), width, height};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character, strlen(character)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeString(char x, char y, char font, short int color, char width, char height, const char * character){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x53, x, y, font, msb(color), lsb(color), width, height};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character, strlen(character)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::placeString(char column, char row, char font, short int color, const char * character){
	#if debug
		fprintf(stderr,"placeString(const char *)\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	//char cmd[] = {0x53, column, row, font, msb(color), lsb(color)};
	char cmd[] = {'s', column, row, font, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(::write(portHandle, character, strlen(character)) <= 0)
		return 0;
	char terminator = 0;
	if( ::write(portHandle, &terminator, 1) <= 0 )
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::placeText(char character, char column, char row, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x54, character, column, row, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::setBacklight(char mode){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x59, 0x00, mode};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::setDisplay(bool mode){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'Y', 0x01, mode};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::setContrast(char value){
	#if debug
		fprintf(stderr,"setContrast(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'Y', 0x02, value};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::powerUp(void){
	#if debug
		fprintf(stderr,"powerUp(\n");
	#endif
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'Y', 0x03, 0x01};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::powerDown(void){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'Y', 0x03, 0x00};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::setOrientation(char mode){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'Y', 0x04, mode};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	int temp=screeninfo.xres;
	screeninfo.xres=screeninfo.yres;
	screeninfo.yres=temp;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::writeRegister(char reg, char data, char data2){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'$', 'W', reg, data, data2};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
/*bool _4DModule::writeRegister(char reg, char data){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x24, 0x57, reg, data};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}*/
bool _4DModule::writeRegister(bool mode, char data){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x24, 0x57, mode, data};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::displayScroll(bool enable, bool direction, char speed){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char en_cmd[] = {0x24, 0x53, 0x00, enable};
	if(::write(portHandle, &en_cmd, sizeof(en_cmd)) <= 0)
		return 0;
	char dir_cmd[] = {0x24, 0x53, 0x01, direction};
	if(::write(portHandle, &dir_cmd, sizeof(dir_cmd)) <= 0)
		return 0;
	char sp_cmd[] = {0x24, 0x53, 0x02, speed};
	if(::write(portHandle, &sp_cmd, sizeof(sp_cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::displayScroll(bool enable, char speed){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char en_cmd[] = {0x24, 0x53, 0x00, enable};
	if(::write(portHandle, &en_cmd, sizeof(en_cmd)) <= 0)
		return 0;
	char sp_cmd[] = {0x24, 0x53, 0x02, speed};
	if(::write(portHandle, &sp_cmd, sizeof(sp_cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::dimScreen(char x, char y, char width, char height){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x24, 0x44, x, y, width, height};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::setFloatPointerParams(bool mode, bool appearance, bool size, short int color){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x24, 0x50, mode, appearance, size, msb(color), lsb(color)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
bool _4DModule::moveFloatPointer(char x, short int y){
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {0x24, 0x4D, x, y};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}
std::vector<int>  _4DModule::getFloatPointerCoord(void){
	std::vector<int> retVal;
// 	retVal.at(0) = -1;				//error
	if( ( this->isOpen() ) <= 0 )
		return retVal;
	int			temp;
	char cmd[] = {0x24, 0x43};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return retVal;
	if(::read(portHandle, &temp, 3) <= 0)
		return retVal;
	retVal.at(0) = address_hi(temp);		//x
	retVal.at(1) = address_mid(temp);		//y(msb)
	retVal.at(2) = address_lo(temp);		//y(lsb)
	return retVal;
}
bool _4DModule::vsyncLock(void){}
bool _4DModule::uSDInit(void) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'i'};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

std::vector<unsigned char> _4DModule::uSDReadSector(int address) {
	std::vector<unsigned char> data(512);
	if( ( this->isOpen() ) <= 0 )
		return data;
	char cmd[] = {'@', 'R', address_hi(address), address_mid(address), address_lo(address)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return data;
	char answer[1];
	for(int i=0; i<512; i++) {
		if(::read(portHandle, answer, 1) <= 0)
			return data;
		else data.at(i)=answer[0];
	}
	return data;
}

bool _4DModule::uSDWriteSector(int address, std::vector<unsigned char> * data) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'W', address_hi(address), address_mid(address), address_lo(address)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	for(int i=0; i<512; i++)
		if(::write(portHandle, &data->at(i), 1) <= 0)
			return false;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return true;
}

unsigned char _4DModule::uSDReadByte(void) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'r'};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	char answer[1];
	if(::read(portHandle, answer, 1) <= 0)
		return 0;
	return answer[0];
}

bool _4DModule::uSDWriteByte(char byte) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'w', byte};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::uSDSetAddress(int address) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'A', address_mmsb(address), address_mlsb(address), address_lmsb(address), address_llsb(address)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::uSDCopyScreen(char x, char y, char width, char height, int address) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'C', x, y, width, height, address_hi(address), address_mid(address), address_lo(address)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::uSDDisplayImage(char x, char y, char width, char height, char colorMode, int address) {
	if( ( this->isOpen() ) <= 0 )
		return 0;
	char cmd[] = {'@', 'I', x, y, width, height, colorMode, address_hi(address), address_mid(address), address_lo(address)};
	if(::write(portHandle, &cmd, sizeof(cmd)) <= 0)
		return 0;
	if(screeninfo.writeWaitForACK)
		return __getACK();
	return 1;
}

bool _4DModule::uSDDisplayObject(int address){}
bool _4DModule::uSDRun(int address){}
bool _4DModule::uSDExit(void){}
/*bool _4DModule::fmDisplayImage(int address){}
bool _4DModule::fmDisplayObject(char x, char y, int address){}
bool _4DModule::fmRun(void){}
bool _4DModule::fmExit(void){}
bool _4DModule::fmWrite(short int pageNumber, std::vector<char> * data){}
int  _4DModule::fmRead(short int pageNumber){}
bool _4DModule::fmErase(void){}
int  _4DModule::vmRead(char page){}
bool _4DModule::vmWrite(char page){}
bool _4DModule::vmDisplay(char page){}*/

int RGB_to16(int red, int grn, int blu) {
	return ( ((red) << 11) | ((grn) << 5) | (blu) );
}

int RGB_24to16(int red, int grn, int blu) {
	return ( ((red >> 3) << 11) | ((grn >> 2) << 5) | (blu >> 3) );
}
