/*
Original by Julien
Heavily modified 2008 by Sven Killig <sven@killig.de>
*/

#include <string>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include "4D.h"
#include "shakira.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <pthread.h>
#include <termios.h>
#include <poll.h>

#include <fstream>
#include <math.h>
using namespace std;

int RGB_24to8(int red, int grn, int blu) {
	return ( ((red >> 5) << 5) | ((grn >> 5) << 2) | (blu >> 6) );
}


#define UNDEFINED -1
#define RETURN_HSV(h, s, v) {HSV.H = h; HSV.S = s; HSV.V = v; return HSV;}
#define RETURN_RGB(r, g, b) {RGB.R = r; RGB.G = g; RGB.B = b; return RGB;}
typedef struct {float R, G, B;} RGBType;
typedef struct {float H, S, V;} HSVType;
RGBType HSV_to_RGB( HSVType HSV ) {
 // H is given on [0, 6] or UNDEFINED. S and V are given on [0, 1].
 // RGB are each returned on [0, 1].
 float h = HSV.H, s = HSV.S, v = HSV.V, m, n, f;
 int i;
 RGBType RGB;

 if (h == UNDEFINED) RETURN_RGB(v, v, v);
 i = floor(h);
 f = h - i;
 if ( !(i&1) ) f = 1 - f; // if i is even
 m = v * (1 - s);
 n = v * (1 - s * f);
 switch (i) {
  case 6:
  case 0: RETURN_RGB(v, n, m);
  case 1: RETURN_RGB(n, v, m);
  case 2: RETURN_RGB(m, v, n)
  case 3: RETURN_RGB(m, n, v);
  case 4: RETURN_RGB(n, m, v);
  case 5: RETURN_RGB(v, m, n);
 }
}



int main(int argc, char *argv[])
 {
	if(argc < 3){
		fprintf(stderr, "%s <serial port> <bps (PICASO only 9600)> <command> <parameters>\n\n", argv[0]);
		fprintf(stderr, "reset\n");
		fprintf(stderr, "powerUp\n");
		fprintf(stderr, "moddown\n");
		fprintf(stderr, "E\n");
		fprintf(stderr, "displayScreenInfo\n");
		fprintf(stderr, "F size\n");
		fprintf(stderr, "B R G B\n");
		fprintf(stderr, "setContrast value\n");
		fprintf(stderr, "s column row font R G B string\n");
		fprintf(stderr, "T char, column, row, R G B\n");
		fprintf(stderr, "O\n");
		fprintf(stderr, "_S: scroll\n");
		fprintf(stderr, "dump: send 'U' 160*128*2 times\n");
		fprintf(stderr, "charmap column row font\n");
		fprintf(stderr, "A char d1 d2 d3 d4 d5 d6 d7 d8\n");
		fprintf(stderr, "p\n");
		fprintf(stderr, "D column row char R G B\n");
		fprintf(stderr, "recs\n");
		fprintf(stderr, "sd: show uSD content\n");
		fprintf(stderr, "read sector\n");
		fprintf(stderr, "font2sd startSector width height startASCII skipHeader\n");
		fprintf(stderr, "lines\n");
		fprintf(stderr, "points\n");
		fprintf(stderr, "circles\n");
		fprintf(stderr, "colours colorMode\n");
		fprintf(stderr, "graph\n");
		fprintf(stderr, "calendar font\n");
		fprintf(stderr, "firmware filename: load firmware (tested on GOLDELOX)\n");
		fprintf(stderr, "4dgl filename: load GOLDELOX 4DGL program\n");
		exit(0);
	}
	_4DModule	uoled128gmd1;
	var_screeninfo	myscreen;
	std::cout << "modconfig(" << argv[1] << ", " << argv[2] << "): " << uoled128gmd1.modconfig(argv[1], atoi(argv[2])) << "\n";
	if(strcmp(argv[3], "reset")==0) {
		uoled128gmd1.reset();
		exit(0);
	}
	if(uoled128gmd1.modup() == 0) {
		std::cout << "modup() == 0!\n";
		//exit(0);
	}
	uoled128gmd1.setTimeout(500);
	myscreen = uoled128gmd1.getScreenInfo();
        std::cout << "Device id = " << myscreen.id << "\n";
        std::cout << "Hardware version = " << myscreen.hardwareVersion << "\n";
        std::cout << "Firmware version = " << myscreen.firmwareVersion << "\n";
        std::cout << "Resolution : " << myscreen.xres << "x" << myscreen.yres << "\n";
	if(strcmp(argv[3], "powerUp")==0) {
		std::cout << "powerUp()\n";
		uoled128gmd1.powerUp();
	} else if(strcmp(argv[3], "moddown")==0) {
		std::cout << "moddown()\n";
		uoled128gmd1.moddown();
	} else if(strcmp(argv[3], "E")==0) {
	 	uoled128gmd1.eraseScreen();
	} else if(strcmp(argv[3], "displayScreenInfo")==0) {
		/*uoled128gmd1.writeRegister(0x00, 0xA0);
		uoled128gmd1.writeRegister(0x01, 0x75);*/
	 	uoled128gmd1.displayScreenInfo();
	} else if(strcmp(argv[3], "F")==0) {
		uoled128gmd1.setFontSize(atoi(argv[4]));
	} else if(strcmp(argv[3], "B")==0) {
		uoled128gmd1.setBackground(RGB_to16(atoi(argv[4]),atoi(argv[5]),atoi(argv[6])));
	} else if(strcmp(argv[3], "setContrast")==0) {
		uoled128gmd1.setContrast(atoi(argv[4]));
	} else if(strcmp(argv[3], "s")==0) {
		uoled128gmd1.placeString(atoi(argv[4]),atoi(argv[5]),atoi(argv[6]),RGB_to16(atoi(argv[7]),atoi(argv[8]),atoi(argv[9])),argv[10]);
	} else if(strcmp(argv[3], "T")==0) {
		uoled128gmd1.placeText(argv[4][0],atoi(argv[5]),atoi(argv[6]),RGB_to16(atoi(argv[7]),atoi(argv[8]),atoi(argv[9])));
//uoled128gmd1.placeText(128,atoi(argv[5]),atoi(argv[6]),atoi(argv[7]));
	} else if(strcmp(argv[3], "O")==0) {
		uoled128gmd1.setTextOpacity(atoi(argv[4]));
	} else if(strcmp(argv[3], "_S")==0) {
		uoled128gmd1.displayScroll(atoi(argv[4]),atoi(argv[5]));
	} else if(strcmp(argv[3], "dump")==0) {
		/*char dump[160*128*2];
		for(int i=0;i<sizeof(dump);i++) dump[i]='U';
		uoled128gmd1.write(dump, sizeof(dump));*/
		for(int i=0;i<160*128*2;i++) uoled128gmd1.initialise();
	} else if(strcmp(argv[3], "charmap")==0) {
		char charmap[127];
		for(int i=1;i<=127;i++) charmap[i-1]=i;
		uoled128gmd1.placeString(atoi(argv[4]), atoi(argv[5]), atoi(argv[6]), 65535, charmap);
	} else if(strcmp(argv[3], "A")==0) {
		std::vector<unsigned char> unvecteur(8);
		for(int i=0;i<unvecteur.size();i++) unvecteur.at(i)=atoi(argv[5+i]);
//		uoled128gmd1.addCharacter(atoi(argv[4]),&unvecteur);
for(int i=0;i<32;i++) uoled128gmd1.addCharacter(i,&unvecteur);
	} else if(strcmp(argv[3], "p")==0) {
		uoled128gmd1.setPenSize(atoi(argv[4]));
	} else if(strcmp(argv[3], "D")==0) {
		uoled128gmd1.displayCharacter(atoi(argv[4]),atoi(argv[5]),atoi(argv[6]),RGB_to16(atoi(argv[7]),atoi(argv[8]),atoi(argv[9])));
	} else if(strcmp(argv[3], "recs")==0) {
		while(true) {
			unsigned short int x1=rand()%myscreen.xres,y1=rand()%myscreen.yres;
			unsigned short int x2=rand()%myscreen.xres,y2=rand()%myscreen.yres;
			uoled128gmd1.drawRectangle(x1,y1,x2,y2,rand()%65535);
		}

	} else if(strcmp(argv[3], "sd")==0) {
		uoled128gmd1.uSDInit();
		uoled128gmd1.setTimeout(5000);
		unsigned short int w=myscreen.xres,h=myscreen.yres;
		unsigned char sectors=w*h*2/512+1;
		char str [80];
		int i=0;
		while(i<16777216) {
			for(int y=0; y<h; y+=h) {
				for(int x=0; x<w; x+=w) {
					uoled128gmd1.uSDDisplayImage(x*w,y*w,w,h,16,i++*sectors);
					scanf("%s",str);
				}
			}
		}

	} else if(strcmp(argv[3], "read")==0) {
		std::vector<unsigned char> unvecteur=uoled128gmd1.uSDReadSector(atoi(argv[4]));
		for(int i=0;i<unvecteur.size();i++) {
			printf("%i=%02X ",i,unvecteur.at(i));
		}
		printf("\n");

	} else if(strcmp(argv[3], "font2sd")==0) {
		unsigned short int w=myscreen.xres,h=myscreen.yres;
		int sector;
		int startSector=atoi(argv[4]);
		unsigned char width=atoi(argv[5]);
		unsigned char height=atoi(argv[6]);
		unsigned char startASCII=atoi(argv[7]);
		bool skipHeader=atoi(argv[8]);

		/*uoled128gmd1.uSDSetAddress(startSector*512);
		uoled128gmd1.uSDWriteByte(width);
		uoled128gmd1.uSDSetAddress(startSector*512+1);
		uoled128gmd1.uSDWriteByte(height);*/

		if(!skipHeader) {
			std::vector<unsigned char> unvecteur(512);
			unvecteur.at(0)=width;
			unvecteur.at(1)=height;
			for(int i=2; i<512; i++) {
				unvecteur.at(i)=0;
			}
			uoled128gmd1.uSDWriteSector(startSector,&unvecteur);
			startSector++;
		}

		/*unsigned char startChar=atoi(argv[8]);
		unsigned char quantity=atoi(argv[9]);*/
		unsigned char sectors=ceil((double)width*(double)height*2.0/512.0);
		printf("sectors=%u\n",(unsigned int)sectors);

		uoled128gmd1.setPenSize(0);
		unsigned char c=0;
		for(int y=0; y<h-height; y+=height) {
			for(int x=0; x<w-width; x+=width) {
				sector=startSector+(startASCII+c)*sectors;
				printf("sector=%u\n",(unsigned int)sector);
				uoled128gmd1.uSDCopyScreen(x,y,width,height,sector);
				uoled128gmd1.drawRectangle(x,y,x+width-1,y+height-1,65535);
				uoled128gmd1.drawRectangle(x,y,x+width-1,y+height-1,0);
				c++;
			}
		}

	} else if(strcmp(argv[3], "lines")==0) {
		while(true) {
			unsigned short int x1=rand()%myscreen.xres,y1=rand()%myscreen.yres;
			unsigned short int x2=rand()%myscreen.xres,y2=rand()%myscreen.yres;
			uoled128gmd1.drawLine(x1,y1,x2,y2,rand()%65535);
		}
	/*} else if(strcmp(argv[3], "clock")==0) {
		while(true) {
			unsigned char x1=rand()%160,y1=rand()%128;
			unsigned char x2=rand()%160,y2=rand()%128;
			uoled128gmd1.drawLine((char)x1,(char)y1,(char)x2,(char)y2,rand()%65535);
			sleep(100);
		}*/
	} else if(strcmp(argv[3], "points")==0) {
		while(true) {
			unsigned short int x1=rand()%myscreen.xres,y1=rand()%myscreen.yres;
			uoled128gmd1.putPixel(x1,y1,rand()%65535);
		}
	/*} else if(strcmp(argv[3], "points128")==0) {
		unsigned char i=0;
		char cmd[1];
		while(i<=127) {
		  //uoled128gmd1.putPixel(i,127-i++,65535);
		  cmd[0]='P';
		  uoled128gmd1.write(cmd, sizeof(cmd));
		  sleep(1);
		  cmd[0]=i;
		  uoled128gmd1.write(cmd, sizeof(cmd));
		  sleep(1);
		  cmd[0]=127-i++;
		  uoled128gmd1.write(cmd, sizeof(cmd));
		  sleep(1);
		  //uoled128gmd1.write(cmd, sizeof(cmd));
		  cmd[0]=255;
		  uoled128gmd1.write(cmd, sizeof(cmd));
		  sleep(1);
		  //uoled128gmd1.write(cmd, sizeof(cmd));
		  //uoled128gmd1.write(cmd, sizeof(cmd));
		  uoled128gmd1.writeWaitForACK(cmd, sizeof(cmd));
		}*/
	} else if(strcmp(argv[3], "circles")==0) {
		while(true) {
			unsigned short int x1=rand()%myscreen.xres,y1=rand()%myscreen.yres,r=rand()%(myscreen.yres-y1);
			uoled128gmd1.drawCircle(x1,y1,r,rand()%65535);
		}
	} else if(strcmp(argv[3], "colours")==0) {
		//sleep(1);
		unsigned char width=myscreen.xres,colorMode=atoi(argv[4]);
		unsigned short int height=myscreen.yres;
		printf("width = %i\n",width);
		printf("height = %i\n",height);
		std::vector<unsigned short int> image(width*height);
		/*float f=pow(2,colorMode)/(width*height);
		for(int i=0; i<width*height; i++) {
			image.at(i) = i*f;
		}*/

		RGBType rgb;
		HSVType hsv;
		float yf=1.0f/(float)height*2.0f;
		float xf=6.0f/(float)width;
		for(int y=0; y<height; y++) {
			for(int x=0; x<width; x++) {
				hsv.H=(float)x*xf;
				if(y<height/2) {
					hsv.S=1.0f;
					hsv.V=(float)y*yf;
				} else {
					hsv.S=1.0f-(float)((float)y-(float)height/2.0f)*yf;
					hsv.V=1.0f;
				}
				rgb=HSV_to_RGB(hsv);
				if(colorMode==8) image.at(y*width+x) = RGB_24to8(rgb.R*255,rgb.G*255,rgb.B*255);
				else image.at(y*width+x) = RGB_24to16(rgb.R*255,rgb.G*255,rgb.B*255);
			}
		}

		uoled128gmd1.displayImage(0,0,width,height,colorMode,&image);

	} else if(strcmp(argv[3], "graph")==0) {
		unsigned short int w=myscreen.xres,h=myscreen.yres;
		unsigned char i=0;
		unsigned char v;
		RGBType rgb;
		HSVType hsv;
		float xf=6.0f/(float)h;
		while(true) {
			//v=rand()%127;
			v=sin((i%h-1)*3.0f/(float)h)*(float)h;
			hsv.H=(float)v*xf;
			hsv.S=1.0f;
			hsv.V=1.0f;
			rgb=HSV_to_RGB(hsv);
			uoled128gmd1.copyPaste((char)1, (char)0, (char)0, (char)0, (char)w-1, (char)h-1);
			uoled128gmd1.drawLine((char)w-1, (char)h-1, (char)w-1, (char)(h-1-v), RGB_24to16(rgb.R*255,rgb.G*255,rgb.B*255));
			uoled128gmd1.drawLine((char)w-1, (char)(h-1-v+1), (char)w-1, (char)0, 0);
			i++;
		}

	} else if(strcmp(argv[3], "calendar")==0) {
	  int font=atoi(argv[4]);
          int col=0;
          int row=0;
	  int color;
          int daysInMonth[]={31,28,31,30,31,30,31,31,30,31,30,31};
          time_t rawtime;
          struct tm * timeinfo;
          time ( &rawtime );
          timeinfo = localtime ( &rawtime );
          char buffer[80];
          strftime(buffer,80,"%B %Y",timeinfo);
          uoled128gmd1.placeString(col,row,font,65535,buffer);
          row+=3;

          uoled128gmd1.placeString(col,row,font,RGB_24to16(255,255,255),"Mo Di Mi Do Fr");
          col+=15;
          uoled128gmd1.placeString(col,row,font,RGB_24to16(127,127,127),"Sa So");
          row+=2;
          int weekday=timeinfo->tm_wday;
          int today=timeinfo->tm_mday;
          int last=daysInMonth[timeinfo->tm_mon];
          col=(today-1+7-weekday)%7;
          int curr=1;
          char buf[3];
          while(curr<=last) {
            sprintf(buf,"%2d",curr);
            if(curr==today)
              color=RGB_24to16(255,0,0);
            else {
              if(col>4) color=RGB_24to16(127,127,127);
              else color=RGB_24to16(255,255,255);
            }
            uoled128gmd1.placeString(col*3,row,font,color,buf);
            /*col++;
            if(col==7) {*/
            if(!(col=++col%7)) {
              //col=0;
              row+=2;
            }
            curr++;
          }

	} else if(strcmp(argv[3], "firmware")==0) {
		ifstream p_firmwareName (argv[4], ios::in|ios::binary|ios::ate);
		if (!p_firmwareName.is_open()) {
			std::cout << "Error opening file !\n";
			exit(0);
		}
		char cmd[] = { '4','D','G','L' };
		char rec[] = { 0, 0, 0, 0 };				//Receiver buffer
		char erase[] = { 'e' };					//Erase module command
		char eraseFlash[] = { 'f' };				//Erase flash command
		char inbuff = 0;					//Little buffer to retrieve module's answer
		int total = (int)p_firmwareName.tellg();			//Number of bytes to be transferred
		//char trials = 50;					//Sets the timeout of the functions: 50 * screeninfo.timeout = 50 * 100ms = 5s
		char trials = 100;
		int nchunks = total / 256;				//Number of chunks (256 bytes) in the file.
		int checksum;						//The verification number
		char * sector = new char[260];				//260 bytes are transferred; found experimentally by looking at PmmCLoader's dump.
		if(total % 256) {					//We don't support not rounded file yet. TODO padd the file if not 256 round.
			std::cout << "ERROR FILE\n";
			exit(0);
		}
		//Send some info to the user about the progress of the programmation
		std::cout << "Initialising device...\n";
		uoled128gmd1.reset();								//Make sure the module is with us
uoled128gmd1.setTimeout(100);
		while( (rec[0]+rec[1]+rec[2]+rec[3] != '4'+'d'+'g'+'l') && trials != 0) {
			if( uoled128gmd1.write(cmd, sizeof(cmd)) <= 0 ) {
				std::cout << "Error while initialising device.1\n";
				exit(0);
			}
			uoled128gmd1.readTimeout(rec, sizeof(rec));
			trials--;
		}
		if(trials <= 0){								//timeout
			std::cout << "Error while initialising device.2\n";
			exit(0);
		}
		std::cout << "Erasing module...\n";
		//This avoids us to set a loop.
		uoled128gmd1.setTimeout(5000);
		if( (uoled128gmd1.writeWaitForACK(erase, sizeof(erase))) <= 0 ) {
			std::cout << "Could not get erase ACK.\n";
			exit(0);
		}
		if( (uoled128gmd1.readTimeout(&inbuff, sizeof(inbuff))) <= 0 ){			//timeout
			std::cout << "Error while erasing device.\n";
			exit(0);
		}
		if(inbuff != 'E')
			exit(0);
		//Restore the timeout for responsiveness
		uoled128gmd1.setTimeout(100);
		std::cout << "Programming";
		sector[0] = 'p';								//The programming command
		for(int i=(nchunks - 1); i>=0; i--){						//Looks like it needs to be programmed in a backward fashion
			std::cout << ".";
			checksum = 0;
			sector[1] = msb(i);							//Tells the chip about our position in the file
			sector[2] = lsb(i);
			p_firmwareName.seekg(i*256);						//This could have been greatly simplified with QDataStream
			p_firmwareName.read((sector+3),256);					//if it did not need to be read the file in a backward fashion.
			for(int j=0; j < 256; j++){
				checksum += sector[3+j];					//Calculate the checksum
			}
			sector[259] = (char)( (-checksum) & 0xFF ); 				//This one was hard to find from the dump
			if( uoled128gmd1.writeWaitForACK(sector, 260) <= 0 ) {
				std::cout << "\nError while writing to device.\n";
				exit(0);
			}
		}
		std::cout << "\nErasing flash...\n";
		uoled128gmd1.setTimeout(5000);
		if( uoled128gmd1.writeWaitForACK(eraseFlash, sizeof(eraseFlash)) <= 0 ) {
			std::cout << "Could not get run ACK.\n";
			exit(0);
		}
		uoled128gmd1.setTimeout(100);
		uoled128gmd1.reset();
		std::cout << "Transfer of the firmware terminated.\n";

	} else if(strcmp(argv[3], "4dgl")==0) {
		ifstream p_firmwareName (argv[4], ios::in|ios::binary|ios::ate);
		if (!p_firmwareName.is_open()) {
			std::cout << "Error opening file !\n";
			exit(0);
		}
		char cmd[] = { '4','d','g','l' };
		char load[] = { 'L', 0 };
		char inbuff = 0;					//Little buffer to retrieve module's answer
		int total = (int)p_firmwareName.tellg()-0x10;			//Number of bytes to be transferred
		//char trials = 50;					//Sets the timeout of the functions: 50 * screeninfo.timeout = 50 * 100ms = 5s
		char trials = 100;					//Sets the timeout of the functions: 50 * screeninfo.timeout = 50 * 100ms = 5s
		char chunksize = 64;
		int nchunks = ceil((double)total / (double)chunksize);				//Number of chunks in the file.
		int checksum;						//The verification number
		char * sector = new char[chunksize+1];				//found experimentally by looking at PmmCLoader's dump.
		//Send some info to the user about the progress of the programmation
		std::cout << "Initialising device...\n";
		uoled128gmd1.reset();								//Make sure the module is with us
		while( (inbuff != 'G') && trials != 0) {
			if( uoled128gmd1.write(cmd, sizeof(cmd)) <= 0 ) {
				std::cout << "Error while initialising device.1\n";
				exit(0);
			}
			uoled128gmd1.readTimeout(&inbuff, sizeof(inbuff));
			trials--;
		}
		if(trials <= 0){								//timeout
			std::cout << "Error while initialising device.2\n";
			exit(0);
		}

		std::cout << "Loading...\n";
		//This avoids us to set a loop.
		uoled128gmd1.setTimeout(5000);
		load[1]=nchunks;
		if( uoled128gmd1.writeWaitForACK(load, sizeof(load)) <= 0 ) {
			std::cout << "Could not get load ACK.\n";
			exit(0);
		}
		//Restore the timeout for responsiveness
		uoled128gmd1.setTimeout(100);
		std::cout << "Programming";
		for(int i=0; i<=nchunks - 1; i++){
			std::cout << ".";
			checksum = 0;
			p_firmwareName.seekg(0x10 + i*chunksize);
			p_firmwareName.read(sector,chunksize);
			for(int j=0; j < chunksize; j++){
				checksum += sector[j];					//Calculate the checksum
			}
			sector[chunksize] = (char)( (-checksum) & 0xFF ); 				//This one was hard to find from the dump
			if( uoled128gmd1.write(sector, chunksize+1) <= 0 ) {
				std::cout << "\nError while writing to device.\n";
				exit(0);
			}
          		if( (uoled128gmd1.readTimeout(&inbuff, sizeof(inbuff))) <= 0 ){			//timeout
          			std::cout << "Error at " << i << "\n";
          			if(inbuff!=0) std::cout << "(inbuff!=0)\n";
          			exit(0);
          		}
		}
		if(uoled128gmd1.__getACK()!=1) {
			std::cout << "Could not get end ACK.\n";
			exit(0);
		}
		uoled128gmd1.setTimeout(100);
		uoled128gmd1.reset();
		std::cout << "\nTransfer of the 4DGL program finished.\n";
	} else {
		std::cout << "Unknown command!\n";
	}
/*string s;
getline(cin, s);*/
	exit(0);
	return 0;
 }
